"""Declarations for Python renderings of basic JavaScript types."""

from __future__ import annotations

from datetime import datetime
from typing import (
    Union,
)


class JSUndefinedType:
    """The JavaScript undefined type.

    Where JavaScript null is represented as None, undefined is represented as this
    type."""

    def __bool__(self) -> bool:
        return False

    def __repr__(self) -> str:
        return "JSUndefined"


JSUndefined = JSUndefinedType()


class JSObject:
    pass


PythonJSConvertedTypes = Union[
    None,
    JSUndefinedType,
    bool,
    int,
    float,
    str,
    JSObject,
    datetime,
    memoryview,
]


class MiniRacerBaseException(Exception):  # noqa: N818
    """Base MiniRacer exception."""


class JSEvalException(MiniRacerBaseException):
    """JavaScript could not be executed."""


class JSTimeoutException(JSEvalException):
    """JavaScript execution timed out."""

    def __init__(self) -> None:
        super().__init__("JavaScript was terminated by timeout")



from gc import collect
from time import sleep, time

import pytest


class GcCheck:
    @staticmethod
    def check(mr):
        """Test helper for garbage collection.

        PyMiniRacer does somewhat tricky things with object lifecycle management
        (basically, various __del__ methods, backed by C++ keeping its own track of
        all allocated objects). This is a somewhat kludgey test helper to verify
        those tricks are working.

        The Python gc doesn't seem particularly deterministic, so we do 2 collects
        and a sleep here to reduce the flake rate.
        """
        start = time()
        while time() - start < 5 and mr._ctx.value_count() != 0:  # noqa: SLF001
            collect()
            sleep(0.05)

        assert mr._ctx.value_count() == 0  # noqa: SLF001


@pytest.fixture
def gc_check():
    return GcCheck



"""Tests JSFunctions"""

import pytest
from py_mini_racer import (
    JSEvalException,
    JSTimeoutException,
    MiniRacer,
)


def test_function(gc_check):
    mr = MiniRacer()
    func = mr.eval("(a) => a")
    assert func(42) == 42
    arr = mr.eval("[41, 42]")
    assert list(func(arr)) == list(func(arr))
    thing = mr.eval(
        """\
class Thing {
    constructor(a) {
        this.blob = a;
    }

    stuff(extra) {
        return this.blob + extra;
    }
}
new Thing('start');
"""
    )
    stuff = thing["stuff"]
    assert stuff("end", this=thing) == "startend"

    del func, arr, thing, stuff
    gc_check.check(mr)


def test_exceptions(gc_check):
    mr = MiniRacer()
    func = mr.eval(
        """\
function func(a, b, c) {
    throw new Error('asdf');
}
func
"""
    )

    with pytest.raises(JSEvalException) as exc_info:
        func()

    assert (
        exc_info.value.args[0]
        == """\
<anonymous>:2: Error: asdf
    throw new Error('asdf');
    ^

Error: asdf
    at func (<anonymous>:2:11)
"""
    )

    del func, exc_info
    gc_check.check(mr)


def test_timeout(gc_check):
    mr = MiniRacer()
    func = mr.eval("() => { while(1) { } }")
    with pytest.raises(JSTimeoutException) as exc_info:
        func(timeout_sec=1)

    assert exc_info.value.args[0] == "JavaScript was terminated by timeout"

    del func, exc_info
    gc_check.check(mr)



from py_mini_racer import MiniRacer


def test_heap_stats(gc_check):
    mr = MiniRacer()

    assert mr.heap_stats()["used_heap_size"] > 0
    assert mr.heap_stats()["total_heap_size"] > 0

    gc_check.check(mr)



from __future__ import annotations

from threading import Condition
from typing import TYPE_CHECKING

from py_mini_racer._types import JSTimeoutException

if TYPE_CHECKING:
    from py_mini_racer._numeric import Numeric
    from py_mini_racer._types import PythonJSConvertedTypes


class SyncFuture:
    """A blocking synchronization object for function return values.

    This is like asyncio.Future but blocking, or like
    concurrent.futures.Future but without an executor.
    """

    def __init__(self) -> None:
        self._cv = Condition()
        self._settled: bool = False
        self._res: PythonJSConvertedTypes = None
        self._exc: Exception | None = None

    def get(self, *, timeout: Numeric | None = None) -> PythonJSConvertedTypes:
        with self._cv:
            while not self._settled:
                if not self._cv.wait(timeout=timeout):
                    raise JSTimeoutException

            if self._exc:
                raise self._exc
            return self._res

    def set_result(self, res: PythonJSConvertedTypes) -> None:
        with self._cv:
            self._res = res
            self._settled = True
            self._cv.notify()

    def set_exception(self, exc: Exception) -> None:
        with self._cv:
            self._exc = exc
            self._settled = True
            self._cv.notify()



"""Test Python functions exposed as JS."""

from asyncio import gather
from asyncio import run as asyncio_run
from asyncio import sleep as asyncio_sleep
from time import time

import pytest
from py_mini_racer import (
    JSPromise,
    JSPromiseError,
    MiniRacer,
)


def test_basic(gc_check):
    mr = MiniRacer()

    data = []

    async def append(*args):
        data.append(args)
        return "foobar"

    async def run():
        async with mr.wrap_py_function(append) as jsfunc:
            # "Install" our JS function on the global "this" object:
            mr.eval("x => this.func = x")(jsfunc)

            assert await mr.eval("this.func(42)") == "foobar"
            assert await mr.eval('this.func("blah")') == "foobar"

            assert data == [(42,), ("blah",)]

    for _ in range(100):
        data[:] = []
        asyncio_run(run())

    gc_check.check(mr)


def test_exception(gc_check):
    # Test a Python callback which raises exceptions
    mr = MiniRacer()

    data = []

    async def append(*args):
        del args
        boo = "boo"
        raise RuntimeError(boo)

    async def run():
        async with mr.wrap_py_function(append) as jsfunc:
            # "Install" our JS function on the global "this" object:
            mr.eval("x => this.func = x")(jsfunc)

            with pytest.raises(JSPromiseError) as exc_info:
                await mr.eval("this.func(42)")

            assert exc_info.value.args[0].startswith(
                """\
JavaScript rejected promise with reason: Error: Error running Python function:
Traceback (most recent call last):
"""
            )

            assert exc_info.value.args[0].endswith(
                """\

    at <anonymous>:1:6
"""
            )

    for _ in range(100):
        data[:] = []
        asyncio_run(run())

    gc_check.check(mr)


def test_slow(gc_check):
    # Test a Python callback which runs slowly, but is faster in parallel.
    mr = MiniRacer()

    data = []

    async def append(*args):
        await asyncio_sleep(1)
        data.append(args)
        return "foobar"

    async def run():
        async with mr.wrap_py_function(append) as jsfunc:
            # "Install" our JS function on the global "this" object:
            mr.eval("x => this.func = x")(jsfunc)

            pending = [mr.eval("this.func(42)") for _ in range(100)]

            assert await gather(*pending) == ["foobar"] * 100

            assert (
                data
                == [
                    (42,),
                ]
                * 100
            )

    start = time()
    asyncio_run(run())
    # The above should run in just over a second.
    # Just verify it didn't take 100 seconds (i.e., that things didn't execute
    # sequentially):
    assert time() - start < 10

    gc_check.check(mr)


def test_call_on_exit(gc_check):
    """Checks that calls from JS made while we're trying to tear down the wrapped
    function are ignored and don't break anything."""

    mr = MiniRacer()

    data = []

    async def append(*args):
        data.append(args)
        return "foobar"

    async def run():
        async with mr.wrap_py_function(append) as jsfunc:
            # "Install" our JS function on the global "this" object:
            mr.eval("x => this.func = x")(jsfunc)

            # Note: we don't await the promise, meaning we just start a call and never
            # finish it:
            assert isinstance(mr.eval("this.func(42)"), JSPromise)

            # Generally (subject to race conditions) at this point the
            # callback initiated by the above this.func(42) will be half-received:
            # _Context.wrap_py_function.on_called will have gotten the callback, and
            # will have told asyncio to deal with it on the loop thread. We will
            # generally *not* have yet processed the call.
            # After this line, we start tearing down the mr.wrap_py_function context
            # manager, which entails stopping the call processor.
            # Let's make sure we don't fall over ourselves (it's fair to either process
            # the last straggling calls, or ignore them, but make sure we don't hang).

    for _ in range(100):
        data[:] = []
        asyncio_run(run())

    gc_check.check(mr)



"""Test .eval() method"""

from asyncio import run as asyncio_run
from time import sleep, time

import pytest
from py_mini_racer import (
    JSEvalException,
    JSOOMException,
    JSParseException,
    JSPromise,
    JSPromiseError,
    JSSymbol,
    JSTimeoutException,
    JSUndefined,
    MiniRacer,
)


def test_invalid(gc_check):
    mr = MiniRacer()

    with pytest.raises(JSEvalException) as exc_info:
        mr.eval("invalid")

    assert (
        exc_info.value.args[0]
        == """\
<anonymous>:1: ReferenceError: invalid is not defined
invalid
^

ReferenceError: invalid is not defined
    at <anonymous>:1:1
"""
    )

    del exc_info
    gc_check.check(mr)


def test_eval(gc_check):
    mr = MiniRacer()
    assert mr.eval("42") == 42

    gc_check.check(mr)


def test_blank(gc_check):
    mr = MiniRacer()
    assert mr.eval("") is JSUndefined
    assert mr.eval(" ") is JSUndefined
    assert mr.eval("\t") is JSUndefined

    gc_check.check(mr)


def test_global(gc_check):
    mr = MiniRacer()
    mr.eval("var xabc = 22;")
    assert mr.eval("xabc") == 22

    gc_check.check(mr)


def test_fun(gc_check):
    mr = MiniRacer()
    mr.eval("var x = function(y) {return y+1;}")

    assert mr.eval("x(1)") == 2
    assert mr.eval("x(10)") == 11
    assert mr.eval("x(100)") == 101

    gc_check.check(mr)


def test_multiple_ctx(gc_check):
    c1 = MiniRacer()
    c2 = MiniRacer()
    c3 = MiniRacer()

    c1.eval("var x = 1")
    c2.eval("var x = 2")
    c3.eval("var x = 3")
    assert c1.eval("(x)") == 1
    assert c2.eval("(x)") == 2
    assert c3.eval("(x)") == 3

    gc_check.check(c1)
    gc_check.check(c2)
    gc_check.check(c3)


def test_exception_thrown(gc_check):
    mr = MiniRacer()

    js_source = "var f = function() {throw new Error('blah')};"

    mr.eval(js_source)

    with pytest.raises(JSEvalException) as exc_info:
        mr.eval("f()")

    assert (
        exc_info.value.args[0]
        == """\
<anonymous>:1: Error: blah
var f = function() {throw new Error('blah')};
                    ^

Error: blah
    at f (<anonymous>:1:27)
    at <anonymous>:1:1
"""
    )

    del exc_info
    gc_check.check(mr)


def test_string_thrown(gc_check):
    mr = MiniRacer()

    js_source = "var f = function() {throw 'blah'};"

    mr.eval(js_source)

    with pytest.raises(JSEvalException) as exc_info:
        mr.eval("f()")

    # When you throw a plain string (not wrapping it in a `new Error(...)`), you
    # get no backtrace:
    assert (
        exc_info.value.args[0]
        == """\
<anonymous>:1: blah
var f = function() {throw 'blah'};
                    ^
"""
    )

    del exc_info
    gc_check.check(mr)


def test_cannot_parse(gc_check):
    mr = MiniRacer()
    js_source = "var f = function("

    with pytest.raises(JSParseException) as exc_info:
        mr.eval(js_source)

    assert (
        exc_info.value.args[0]
        == """\
<anonymous>:1: SyntaxError: Unexpected end of input
var f = function(
                 ^

SyntaxError: Unexpected end of input
"""
    )

    del exc_info
    gc_check.check(mr)


def test_null_byte(gc_check):
    mr = MiniRacer()

    s = "\x00 my string!"

    # Try return a string including a null byte
    in_val = 'var str = "' + s + '"; str;'
    result = mr.eval(in_val)
    assert result == s

    gc_check.check(mr)


def test_timeout(gc_check):
    timeout = 0.1
    start_time = time()

    mr = MiniRacer()
    with pytest.raises(JSTimeoutException) as exc_info:
        mr.eval("while(1) { }", timeout_sec=timeout)

    duration = time() - start_time
    # Make sure it timed out on time, and allow a giant leeway (because aarch64
    # emulation tests are surprisingly slow!)
    assert timeout <= duration <= timeout + 5

    assert exc_info.value.args[0] == "JavaScript was terminated by timeout"

    del exc_info
    gc_check.check(mr)


def test_timeout_ms(gc_check):
    # Same as above but with the deprecated timeout millisecond argument
    timeout = 0.1
    start_time = time()

    mr = MiniRacer()
    with pytest.raises(JSTimeoutException) as exc_info:
        mr.eval("while(1) { }", timeout=int(timeout * 1000))

    duration = time() - start_time
    # Make sure it timed out on time, and allow a giant leeway (because aarch64
    # emulation tests are surprisingly slow!)
    assert timeout <= duration <= timeout + 5

    assert exc_info.value.args[0] == "JavaScript was terminated by timeout"

    del exc_info
    gc_check.check(mr)


def test_max_memory_soft(gc_check):
    mr = MiniRacer()
    mr.set_soft_memory_limit(100000000)
    mr.set_hard_memory_limit(100000000)
    with pytest.raises(JSOOMException) as exc_info:
        mr.eval(
            """\
let s = 1000;
var a = new Array(s);
a.fill(0);
while(true) {
    s *= 1.1;
    let n = new Array(Math.floor(s));
    n.fill(0);
    a = a.concat(n);
}
"""
        )

    assert mr.was_soft_memory_limit_reached()
    assert mr.was_hard_memory_limit_reached()
    assert exc_info.value.args[0] == "JavaScript memory limit reached"

    del exc_info
    gc_check.check(mr)


def test_max_memory_hard(gc_check):
    mr = MiniRacer()
    mr.set_hard_memory_limit(100000000)
    with pytest.raises(JSOOMException) as exc_info:
        mr.eval(
            """\
let s = 1000;
var a = new Array(s);
a.fill(0);
while(true) {
    s *= 1.1;
    let n = new Array(Math.floor(s));
    n.fill(0);
    a = a.concat(n);
}"""
        )

    assert not mr.was_soft_memory_limit_reached()
    assert mr.was_hard_memory_limit_reached()
    assert exc_info.value.args[0] == "JavaScript memory limit reached"

    del exc_info
    gc_check.check(mr)


def test_max_memory_hard_eval_arg(gc_check):
    # Same as above but passing the argument into the eval method (which is a
    # deprecated thing to do because the parameter is really affine to the
    # MiniRacer object)
    mr = MiniRacer()
    with pytest.raises(JSOOMException) as exc_info:
        mr.eval(
            """\
let s = 1000;
var a = new Array(s);
a.fill(0);
while(true) {
    s *= 1.1;
    let n = new Array(Math.floor(s));
    n.fill(0);
    a = a.concat(n);
}""",
            max_memory=200000000,
        )

    assert exc_info.value.args[0] == "JavaScript memory limit reached"

    del exc_info
    gc_check.check(mr)


def test_symbol(gc_check):
    mr = MiniRacer()
    res = mr.eval("Symbol.toPrimitive")
    assert isinstance(res, JSSymbol)

    del res
    gc_check.check(mr)


def test_microtask(gc_check):
    # PyMiniRacer uses V8 microtasks (things, like certain promise callbacks, which run
    # immediately after an evaluation ends).
    # By default, V8 runs any microtasks before it returns control to PyMiniRacer.
    # Let's test that they actually work.
    # PyMiniRacer does not expose the web standard `window.queueMicrotask` (because it
    # does not expose a `window` to begin with). We can, however, trigger a microtask
    # by triggering one as a side effect of a `then` on a resolved promise:
    mr = MiniRacer()
    assert not mr.eval(
        """
let p = Promise.resolve();

var done = false;

p.then(() => {done = true});

done
"""
    )
    assert mr.eval("done")

    gc_check.check(mr)


def test_longer_microtask(gc_check):
    # Verifies a bug fix wherein failure to set a v8::Isolate::Scope on the message
    # pump thread would otherwise result in a segmentation fault:
    mr = MiniRacer()
    mr.eval(
        """
var done = false;
async function foo() {
    await new Promise((res, rej) => setTimeout(res, 1000));
    for (let i = 0; i < 10000000; i++) { }
    done = true;
}
foo();
"""
    )

    assert not mr.eval("done")
    start = time()
    while time() - start < 10 and not mr.eval("done"):
        sleep(0.1)
    assert mr.eval("done")

    gc_check.check(mr)


def test_polling(gc_check):
    mr = MiniRacer()
    assert not mr.eval(
        """
var done = false;
setTimeout(() => { done = true; }, 1000);
done
"""
    )
    assert not mr.eval("done")
    start = time()
    # Give the 1-second wait 10 seconds to finish. (Emulated aarch64 tests are
    # surprisingly slow!)
    while time() - start < 10 and not mr.eval("done"):
        sleep(0.1)
    assert mr.eval("done")

    gc_check.check(mr)


def test_settimeout(gc_check):
    mr = MiniRacer()
    mr.eval(
        """
var results = [];
let a = setTimeout(() => { results.push("a"); }, 2000);
let b = setTimeout(() => { results.push("b"); }, 3000);
let c = setTimeout(() => { results.push("c"); }, 1000);
let d = setTimeout(() => { results.push("d"); }, 4000);
clearTimeout(b)
"""
    )
    start = time()
    # Give the 1-second wait 10 seconds to finish. (Emulated aarch64 tests are
    # surprisingly slow!)
    while time() - start < 10 and mr.eval("results.length") != 3:
        sleep(0.1)
    assert mr.eval("results.length") == 3
    assert mr.eval("results[0]") == "c"
    assert mr.eval("results[1]") == "a"
    assert mr.eval("results[2]") == "d"

    gc_check.check(mr)


def test_promise_sync(gc_check):
    mr = MiniRacer()
    promise = mr.eval(
        """
new Promise((res, rej) => setTimeout(() => res(42), 1000)); // 1 s timeout
"""
    )
    assert isinstance(promise, JSPromise)
    start = time()
    # Give the 1-second wait 10 seconds to finish. (Emulated aarch64 tests are
    # surprisingly slow!)
    result = promise.get(timeout=10)
    assert time() - start > 0.5
    assert result == 42

    del promise
    gc_check.check(mr)


def test_promise_async(gc_check):
    mr = MiniRacer()

    async def run_test():
        promise = mr.eval(
            """
new Promise((res, rej) => setTimeout(() => res(42), 1000)); // 1 s timeout
"""
        )
        assert isinstance(promise, JSPromise)
        start = time()
        result = await promise
        assert time() - start > 0.5
        # Give the 1-second wait 10 seconds to finish. (Emulated aarch64 tests are
        # surprisingly slow!)
        assert time() - start < 10
        assert result == 42

    asyncio_run(run_test())
    gc_check.check(mr)


def test_resolved_promise_sync(gc_check):
    mr = MiniRacer()
    val = mr.eval("Promise.resolve(6*7)").get()
    assert val == 42

    gc_check.check(mr)


def test_resolved_promise_async(gc_check):
    mr = MiniRacer()

    async def run_test():
        val = await mr.eval("Promise.resolve(6*7)")
        assert val == 42

    asyncio_run(run_test())
    gc_check.check(mr)


def test_rejected_promise_sync(gc_check):
    mr = MiniRacer()
    with pytest.raises(JSPromiseError) as exc_info:
        mr.eval("Promise.reject(new Error('this is an error'))").get()

    assert (
        exc_info.value.args[0]
        == """\
JavaScript rejected promise with reason: Error: this is an error
    at <anonymous>:1:16
"""
    )

    del exc_info
    gc_check.check(mr)


def test_rejected_promise_async(gc_check):
    mr = MiniRacer()

    async def run_test():
        with pytest.raises(JSPromiseError) as exc_info:
            await mr.eval("Promise.reject(new Error('this is an error'))")

        assert (
            exc_info.value.args[0]
            == """\
JavaScript rejected promise with reason: Error: this is an error
    at <anonymous>:1:16
"""
        )

    asyncio_run(run_test())
    gc_check.check(mr)


def test_rejected_promise_sync_stringerror(gc_check):
    mr = MiniRacer()
    with pytest.raises(JSPromiseError) as exc_info:
        mr.eval("Promise.reject('this is a string')").get()

    assert (
        exc_info.value.args[0]
        == """\
JavaScript rejected promise with reason: this is a string
"""
    )

    del exc_info
    gc_check.check(mr)



"""Test loading and executing babel.js"""

from os.path import dirname
from os.path import join as pathjoin

from py_mini_racer import MiniRacer


def test_babel(gc_check):
    mr = MiniRacer()

    path = pathjoin(dirname(__file__), "fixtures/babel.js")
    with open(path, encoding="utf-8") as f:
        babel_source = f.read()
    source = f"""
      var self = this;
      {babel_source}
      babel.eval = function(code) {{
        return eval(babel.transform(code)["code"]);
      }}
    """
    mr.eval(source)
    assert mr.eval("babel.eval(((x) => x * x)(8))") == 64

    gc_check.check(mr)



from traceback import clear_frames

import pytest
from py_mini_racer import JSEvalException, JSUndefined, StrictMiniRacer


def test_basic_int(gc_check):
    mr = StrictMiniRacer()
    assert mr.execute("42") == 42

    gc_check.check(mr)


def test_basic_string(gc_check):
    mr = StrictMiniRacer()
    assert mr.execute('"42"') == "42"

    gc_check.check(mr)


def test_basic_hash(gc_check):
    mr = StrictMiniRacer()
    assert mr.execute("{}") == {}

    gc_check.check(mr)


def test_basic_array(gc_check):
    mr = StrictMiniRacer()
    assert mr.execute("[1, 2, 3]") == [1, 2, 3]

    gc_check.check(mr)


def test_call(gc_check):
    js_func = """var f = function(args) {
        return args.length;
    }"""

    mr = StrictMiniRacer()

    assert mr.eval(js_func) is JSUndefined
    assert mr.call("f", list(range(5))) == 5

    gc_check.check(mr)


def test_message(gc_check):
    mr = StrictMiniRacer()
    with pytest.raises(JSEvalException) as exc_info:
        mr.eval("throw new EvalError('Hello', 'someFile.js', 10);")

    clear_frames(exc_info.tb)
    gc_check.check(mr)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Callable,
)

from py_mini_racer._types import JSUndefined

if TYPE_CHECKING:
    from contextlib import AbstractContextManager

    from py_mini_racer._numeric import Numeric
    from py_mini_racer._objects import (
        JSArray,
        JSFunction,
        JSPromise,
    )
    from py_mini_racer._types import (
        JSEvalException,
        JSObject,
        JSUndefinedType,
        PythonJSConvertedTypes,
    )


class AbstractValueHandle(ABC):
    @property
    @abstractmethod
    def raw(self) -> object:
        pass

    @abstractmethod
    def to_python(self) -> PythonJSConvertedTypes | JSEvalException:
        pass

    @abstractmethod
    def to_python_or_raise(self) -> PythonJSConvertedTypes:
        pass


class AbstractContext(ABC):
    """A Context provides Pythonic wrappers around the MiniRacer C API.

    This is intended for internal usage by py_mini_racer. MiniRacerContext provides
    a further wrapper around this interface.
    """

    @abstractmethod
    def get_identity_hash(self, obj: JSObject) -> int:
        pass

    @abstractmethod
    def get_own_property_names(
        self, obj: JSObject
    ) -> tuple[PythonJSConvertedTypes, ...]:
        pass

    @abstractmethod
    def get_object_item(
        self, obj: JSObject, key: PythonJSConvertedTypes
    ) -> PythonJSConvertedTypes:
        pass

    @abstractmethod
    def set_object_item(
        self, obj: JSObject, key: PythonJSConvertedTypes, val: PythonJSConvertedTypes
    ) -> None:
        pass

    @abstractmethod
    def del_object_item(self, obj: JSObject, key: PythonJSConvertedTypes) -> None:
        pass

    @abstractmethod
    def del_from_array(self, arr: JSArray, index: int) -> None:
        pass

    @abstractmethod
    def array_insert(
        self, arr: JSArray, index: int, new_val: PythonJSConvertedTypes
    ) -> None:
        pass

    @abstractmethod
    def call_function(
        self,
        func: JSFunction,
        *args: PythonJSConvertedTypes,
        this: JSObject | JSUndefinedType = JSUndefined,
        timeout_sec: Numeric | None = None,
    ) -> PythonJSConvertedTypes:
        pass

    @abstractmethod
    def js_callback(
        self, func: Callable[[PythonJSConvertedTypes | JSEvalException], None]
    ) -> AbstractContextManager[JSFunction]:
        pass

    @abstractmethod
    def promise_then(
        self, promise: JSPromise, on_resolved: JSFunction, on_rejected: JSFunction
    ) -> None:
        pass

    @abstractmethod
    def create_intish_val(self, val: int, typ: int) -> AbstractValueHandle:
        pass

    @abstractmethod
    def create_doublish_val(self, val: float, typ: int) -> AbstractValueHandle:
        pass

    @abstractmethod
    def create_string_val(self, val: str, typ: int) -> AbstractValueHandle:
        pass

    @abstractmethod
    def free(self, val_handle: AbstractValueHandle) -> None:
        pass

    @abstractmethod
    def evaluate(
        self,
        code: str,
        timeout_sec: Numeric | None = None,
    ) -> PythonJSConvertedTypes:
        pass
